
#TODO Schema actions
class InvalidSchemaError(Exception):

    def __init__(self, message=""):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message}'


class NrowsConsistencyError(Exception):

    def __init__(self,name, input_nrows, output_nrows):
        self.name = name
        self.value = 100*round(output_nrows/input_nrows,2)
        self.input_nrows = input_nrows
        self.output_nrows = output_nrows
        super().__init__()

    def __str__(self):
        d = f"{self.name} has less rows than expected: {self.output_nrows} out of a maximum of {self.input_nrows} - {self.value}%"
        return d


import logging

from functools import reduce
import pandas as pd
from sklearn import model_selection

from kensu.pandas import DataFrame,Series
from kensu.numpy import ndarray
from kensu.utils.kensu_provider import KensuProvider
from kensu.utils.dsl import mapping_strategies
from kensu.utils.helpers import eventually_report_in_mem


def train_test_split(*arrays, **options):
    kensu = KensuProvider().instance()
    are_all_ok = reduce(lambda x, y: x and (isinstance(y, DataFrame) or isinstance(y, ndarray) or isinstance(y, Series)), arrays, True)

    a = []
    b = []
    for element in arrays:
        if isinstance(element, DataFrame):
            a.append(element.get_df())
            b.append(element.get_df())
            b.append(element.get_df())
        elif isinstance(element, Series):
            a.append(element.get_s())
            b.append(element.get_s())
            b.append(element.get_s())
    arrays = tuple(a)

    results = model_selection.train_test_split(*arrays, **options)



    if are_all_ok:

        for df, df_result in list(zip(b, results)):
            orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref))
            orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, df))

            result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(df_result, kensu.default_physical_location_ref))
            result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, df_result))
            if kensu.mapping:
                if isinstance(df_result,pd.Series):
                    col = df_result.name
                    kensu.add_dependencies_mapping(result_sc.to_guid(), col, orig_sc.to_guid(), col, 'split_train_test')

                else:
                    for col in df_result:
                        kensu.add_dependencies_mapping(result_sc.to_guid(),col,orig_sc.to_guid(),col,'split_train_test')
            else:
                kensu.add_dependency((df, orig_ds, orig_sc), (df_result, result_ds, result_sc),
                               mapping_strategy=mapping_strategies.DIRECT)
    else:
        logging.debug("Unsupported types in : " + ", ".join(map(lambda x: x.__class__.__name__, arrays)))
        print("Unsupported types in : " + ", ".join(map(lambda x: x.__class__.__name__, arrays)))


    new_results = []
    for x in results:
        if isinstance(x,pd.DataFrame):
            new_results.append(DataFrame.using(x))
        elif isinstance(x,pd.Series):
            new_results.append(Series.using(x))
        else:
            new_results.append(x)

    return new_results

train_test_split.__doc__ = model_selection.train_test_split.__doc__


import logging
from hashlib import sha256

import pandas as pd
import numpy as np

import kensu
from kensu.client import *
from kensu.utils.dsl.extractors import ExtractorSupport, get_or_set_rand_location
from kensu.utils.helpers import singleton, save_stats_json, to_datasource


@singleton
class KensuPandasSupport(ExtractorSupport):  # should extends some KensuSupport class

    def is_supporting(self, df):
        return isinstance(df, pd.DataFrame) or isinstance(df, pd.Series)

    def is_machine_learning(self, df):
        return False

    def skip_wr(self, df):
        if isinstance(df, kensu.pandas.DataFrame):
            return df.get_df()
        elif isinstance(df, kensu.pandas.Series):
            return df.get_s()
        else:
            return df

    # return list of FieldDef
    def extract_schema_fields(self, df):
        df = self.skip_wr(df)
        if isinstance(df, pd.DataFrame):
            return [FieldDef(name=str(k), field_type=v.name, nullable=True) for (k, v) in df.dtypes.to_dict().items()]
        elif isinstance(df, pd.Series):
            if df.name is None:
                return [FieldDef(name='value', field_type=str(df.dtypes), nullable=True)]
            else:
                return [FieldDef(name=df.name, field_type=str(df.dtypes), nullable=True)]

    def extract_location(self, df, location):
        
        if location is not None:
            return location
        else:
            df = self.skip_wr(df)
            from kensu.pandas import DataFrame
            if isinstance(df,DataFrame):
                df = df.get_df()
            # FIXME: or should it be added as prop to the ksu wrapper instead?!
            return get_or_set_rand_location(df)

    def extract_format(self, df, fmt):
        if fmt is not None:
            return fmt
        else:
            df = self.skip_wr(df)
            return df.__class__.__name__

    def tk(self, k, k1): return str(k) + '.' + k1

    # return dict of doubles (stats)
    def extract_stats(self, df):
        df = self.skip_wr(df)
        if isinstance(df, pd.DataFrame):
            try:
                df_desc_numbers = df.describe(include=['number']).to_dict().items()
            except:
                df_desc_numbers = {}

            stats_dict = {self.tk(k, k1): v for k, o in df_desc_numbers for k1, v in o.items()}

            #Extract datetime for timeliness
            date_df = df.select_dtypes(['datetime', 'datetimetz'])
            date_dict = {}
            # if not checked, when there's no Datetime types,
            # it would throw `ValueError: Cannot describe a DataFrame without columns`
            if not (date_df.ndim == 2 and date_df.columns.size == 0):
                date_describe = date_df.describe(datetime_is_numeric=False).to_dict()
                for col in date_describe:
                    first = date_describe[col]['first'].timestamp()*1000
                    last = date_describe[col]['last'].timestamp()*1000
                    date_dict[col + '.first'] = first
                    date_dict[col + '.last'] = last
            
            #Extract categories and boolean as categorical series
            cat_names = df.select_dtypes(['category', 'boolean', 'object']).columns
            cat_dict = {}
            for cat_col in cat_names:
                vc = df[cat_col].value_counts()
                vc_dict = vc.to_dict()
                for k, v in vc_dict.items():
                    if not np.isnan(v):
                        cat_dict[self.tk(cat_col, k)] = v
                count_names = ["nrows", "count", "len", "numrows"]
                count_name = None
                for n in count_names:
                    if n in vc_dict:
                        pass
                    else:
                        count_name = n
                        break
                if count_name is None:
                    logging.warning("Because the categorical value has many category names matches 'nrows', 'count', and alike, the count stat is set to 'kensu.nrows' for " + cat_col)
                    count_name = "kensu.nrows"
                cat_dict[self.tk(cat_col, count_name)] = vc.sum()
                
                distinct_count_name = "num_categories"
                if distinct_count_name in vc_dict:
                    logging.warning("Because the categorical value has a category names 'num_categories', the number of categories stat is set to 'kensu.num_categories' for " + cat_col)
                    distinct_count_name = self.tk("kensu", distinct_count_name)
                cat_dict[self.tk(cat_col, distinct_count_name)] = len(vc_dict)
                
                # TODO do we need nulls and such?

            stats_dict = {**stats_dict, **date_dict, **cat_dict}

            #Add missing value computation
            count = len(df)
            stats_dict['nrows'] = count
            for col in df:
                try:
                    stats_dict[col + '.nullrows'] =  count - stats_dict[col + '.count']
                except:
                    try:
                        stats_dict[col + '.nullrows'] = count - stats_dict[col + '.nrows']
                    except:
                        logging.debug('Unable to get NA count for ' + str(col))

            for key, item in stats_dict.copy().items():
                if isinstance(item,str):
                    del stats_dict[key]
                elif np.isnan(item):
                    del stats_dict[key]
                elif isinstance(item, np.number):
                    stats_dict[key] = item.item()
                
            return stats_dict
        elif isinstance(df, pd.Series):
            return {k: v for k, v in df.describe().to_dict().items() if type(v) in [int,float] }

    def extract_data_source(self, df, pl, **kwargs):

        logical_naming = kwargs["logical_naming"] if "logical_naming" in kwargs else None

        df = self.skip_wr(df)
        location = self.extract_location(df, kwargs.get("location"))
        fmt = self.extract_format(df, kwargs.get("format"))

        if location is None or fmt is None:
            raise Exception(
                "cannot report new pandas dataframe without location ({}) a format provided ({})!".format(location, fmt))

        ds_pk = DataSourcePK(location=location, physical_location_ref=pl)
        name = ('/').join(location.split('/')[-2:])

        return to_datasource(ds_pk=ds_pk, format=fmt, location=location, logical_naming=logical_naming, name=name)

    def extract_schema(self, data_source, df):
        df = self.skip_wr(df)
        fields = self.extract_schema_fields(df)
        sc_pk = SchemaPK(data_source.to_ref(), fields=fields)
        schema = Schema(name="schema:"+data_source.name, pk=sc_pk)
        return schema

    def extract_data_source_and_schema(self, df, pl, **kwargs):
        df = self.skip_wr(df)
        ds = self.extract_data_source(df, pl, **kwargs)
        sc = self.extract_schema(ds, df)
        return ds, sc


from kensu.utils.dsl.with_output_builder_element import WithOutputBuilderElement


class ProcessLineageDepsBuilder(object):
    def __init__(self, kensu, lineage_builder):
        self.kensu = kensu
        self.lineage_builder = lineage_builder

        self.input = None
        self.input_data_source = None
        self.input_schema = None
        self.input_fields_name = None

        self.output = None
        self.output_data_source = None
        self.output_schema = None
        self.output_fields_name = None

        self.data = {}
        self.control = {}

    def with_input(self, i):  # input => pandas df for example
        if self.input is None:
            if isinstance(i, tuple):
                value = i[0]
                data_source = i[1]
                schema = i[2]
            else:
                value = i
                kensu = self.builder.kensu
                data_source = kensu.extractors.extract_data_source(value, kensu.default_physical_location_ref)
                schema = kensu.extractors.extract_schema(data_source, value)

            self.input = value
            self.input_data_source = data_source._report()
            self.input_schema = schema._report()
            self.input_fields_name = [f.name for f in self.input_schema.pk.fields]

        return WithOutputBuilderElement(self)



from kensu.utils.kensu_provider import KensuProvider


def remove_ksu_wrapper(arg):
    # FIXME: fix possible import issues if dependency libs not installed...
    from kensu.pandas import DataFrame
    from kensu.pandas import Series
    from kensu.numpy import ndarray

    if isinstance(arg, DataFrame):
        return arg.get_df()
    elif isinstance(arg, Series):
        return arg.get_s()
    elif isinstance(arg, ndarray):
        return arg.get_nd()
    else:
        return arg


def remove_ksu_wrappers(args):
    new_args = []
    for arg in args:
        new_args.append(remove_ksu_wrapper(arg))
    return new_args

def remove_ksu_kwargs_wrappers(kwargs):
    new_kwargs = {}
    for item in kwargs:
        new_kwargs[item]=remove_ksu_wrapper(kwargs[item])
    return new_kwargs

def kensu_wrapper(method):
    def wrapper(*args, **kwargs):
        new_args = remove_ksu_wrappers(args)
        new_kwargs = remove_ksu_kwargs_wrappers(kwargs)

        result = method(*new_args, **new_kwargs)
        return result

    wrapper.__doc__ = method.__doc__
    return wrapper

def kensu_wrapper_and_report(method,obj):

    from kensu.pandas import DataFrame
    from kensu.pandas import Series
    from kensu.numpy import ndarray

    def wrapper(*args, **kwargs):
        new_args = remove_ksu_wrappers(args)
        new_kwargs = remove_ksu_kwargs_wrappers(kwargs)

        for arg in args:
            if isinstance(arg, DataFrame) or isinstance(arg, Series) or isinstance(arg, ndarray) :
                obj.add_inheritance(arg)
        for kwarg in kwargs:
            if isinstance(kwargs[kwarg],list):
                for item in kwargs[kwarg]:
                    if isinstance(item, DataFrame) or isinstance(item, Series) or isinstance(item, ndarray) :
                        obj.add_inheritance(item)
            elif isinstance(kwargs[kwarg], DataFrame) or isinstance(kwargs[kwarg], Series) or isinstance(kwargs[kwarg], ndarray) :
                obj.add_inheritance(kwargs[kwarg])

        result = method(*new_args, **new_kwargs)
        return result

    wrapper.__doc__ = method.__doc__
    return wrapper


import logging

from kensu.client.models import Schema


class KensuDatasourceAndSchema:
    def __init__(self,
                 ksu_ds,
                 ksu_schema,
                 f_get_stats=lambda: None,
                 f_publish_stats = lambda lineage_run_id : {}):
        """
        :param kensu.client.DataSource ksu_ds: Datasource info which was extracted from external system
        :param kensu.client.Schema ksu_schema: Schema info for the datasource
        :param () -> typing.Dict[string, float] f_get_stats: [Optional] a function to retrieve DataStats for this DataSource
        :param (str) -> () f_publish_stats : [Optional] a function publishing the statistics given a lineage run guid
        """

        self.ksu_ds = ksu_ds
        self.ksu_schema = ksu_schema
        # this allows expensive delaying computation of stats until they are needed (if at all)
        self.f_get_stats = f_get_stats
        # this allows remote computation of stats to be published by the remote processor (if at all)
        self.f_publish_stats = f_publish_stats

    def field_names(self):
        return [f.name for f in self.ksu_schema.pk.fields]

    @staticmethod
    def for_path_with_opt_schema(ksu, ds_path, format=None, categories=None, maybe_schema=None, ds_name=None, f_get_stats=lambda: {}):
        from kensu.client import DataSourcePK, DataSource, FieldDef, SchemaPK, Schema
        pl_ref = ksu.UNKNOWN_PHYSICAL_LOCATION.to_ref()
        ds_pk = DataSourcePK(location=ds_path, physical_location_ref=pl_ref)
        ds = DataSource(name=ds_name or ds_path, format=format, categories=categories, pk=ds_pk)
        if maybe_schema is None:
            maybe_schema = [("unknown", "unknown"), ]
        # logging.debug(str(maybe_schema))
        fields = list([
            FieldDef(name=str(name), field_type=str(dtype), nullable=True)
            for (name, dtype) in maybe_schema
        ])
        schema = Schema(name="schema:" + ds.name,
                        pk=SchemaPK(ds.to_ref(), fields=fields))
        return KensuDatasourceAndSchema(ksu_ds=ds, ksu_schema=schema, f_get_stats=f_get_stats)

    def __repr__(self):
        return 'KensuDatasourceAndSchema(ksu_ds={}, ksu_schema={})'.format(str(self.ksu_ds), str(self.ksu_schema))


class ExtDependencyEntry:
    def __init__(
            self,
            input_ds=None,  # type: KensuDatasourceAndSchema
            # {out_column: [input_column1, input_column2, ...]}
            lineage=None  # type: dict[str, list[str]]
    ):
        """
        :param KensuDatasourceAndSchema input_ds:
        :param typing.Dict[str, typing.List[str]]  lineage: format is  `{out_column: [input_column1, input_column2, ...]}`
        """
        self.input_ds = input_ds
        if lineage is None:
            lineage = {}
        self.lineage = lineage

    def __repr__(self):
        return 'ExtDependencyEntry(input_ds={}, lineage={})'.format(str(self.input_ds), str(self.lineage))


def filter_matching_fields(fields, other_fields):
    """return fields which are same as in other_fields list, ignoring the case"""
    other_fields_lowercase = set([f.lower() for f in other_fields])
    return [f for f in fields if f.lower() in other_fields_lowercase]


class GenericComputedInMemDs:
    def __init__(
            self,
            inputs=None,  # type: list[KensuDatasourceAndSchema]
            lineage=None  # type: list[ExtDependencyEntry]
    ):
        if inputs is None and lineage is not None:
            inputs = [item.input_ds for item in lineage]
        self.lineage = lineage or []
        self.inputs = inputs or []

    def __repr__(self):
        return 'GenericComputedInMemDs(lineage={})'.format(str(self.lineage))

    @staticmethod
    def for_direct_or_full_mapping(
            all_inputs,  # type: list[KensuDatasourceAndSchema]
            out_field_names # type: list[str]
                                   ):
        inputs_with_matching_fields = [(i, filter_matching_fields(i.field_names(), out_field_names))
                                       for i in all_inputs
                                       if len(filter_matching_fields(i.field_names(), out_field_names)) > 0]

        global_lineage = []

        if len(inputs_with_matching_fields) > 0:
            # direct lineage only for matching fields/DSes
            for input, matching_input_fields in inputs_with_matching_fields:
                matching_output_fields = filter_matching_fields(out_field_names, matching_input_fields)
                lin_entry = ExtDependencyEntry(
                    input_ds=input,
                    # lineage: dict[str-output field, list[str]-input fields]
                    lineage=dict(
                        [(out_field_name, filter_matching_fields(input.field_names(), [out_field_name]))
                         for out_field_name in matching_output_fields])
                )
                global_lineage.append(lin_entry)
        else:
            # all-to-all lineage between all output fields & all input DSes fields
            for input in all_inputs:
                lin_entry = ExtDependencyEntry(
                    input_ds=input,
                    # lineage: dict[str-output field, list[str]-input fields]
                    lineage=dict(
                        [(out_field_name, [v.name for v in input.field_names()])
                         for out_field_name in out_field_names])
                )
                global_lineage.append(lin_entry)
        return GenericComputedInMemDs(lineage=global_lineage)



    # FIXME: rename register_output_orig_data -> output_is_not_in_mem ?
    def report(self, ksu, df_result, operation_type, report_output=False, register_output_orig_data=False):
        from kensu.utils.helpers import extract_ksu_ds_schema
        from kensu.utils.dsl import mapping_strategies
        for input_ds in self.inputs:
            extract_ksu_ds_schema(ksu, input_ds, report=True, register_orig_data=True)
        # report output (if needed)
        if isinstance(df_result,Schema):
            result_ds = None
            result_schema = df_result._report()
            ksu.real_schema_df[result_schema.to_guid()] = df_result
        else:
            result_ds, result_schema = extract_ksu_ds_schema(ksu, df_result, report=report_output, register_orig_data=register_output_orig_data)
        # register the lineage
        for dep in self.lineage:
            input_ds = dep.input_ds
            # at column level
            if ksu.mapping:
                # lineage = { out_column: [input_column1, ...] }
                for out_col, input_cols in dep.lineage.items():
                    for in_col in input_cols:
                        # FIXME: maybe there's an issue here is that result_schema DO NOT physically exist?
                        ksu.add_dependencies_mapping(guid=result_schema.to_guid(),
                                                     col=str(out_col),
                                                     from_guid=input_ds.ksu_schema.to_guid(),
                                                     from_col=str(in_col),
                                                     type=operation_type)
            # at datasource level (not column lineage info known)
            else:
                i=(input_ds, input_ds.ksu_ds, input_ds.ksu_schema)
                o=(df_result, result_ds, result_schema)
                ksu.add_dependency(i=i,
                                   o=o,
                                   mapping_strategy=mapping_strategies.FULL)


    @staticmethod
    def report_copy_with_opt_schema(src, dest, operation_type, maybe_schema=None, dest_name=None, src_name=None):
        from kensu.utils.kensu_provider import KensuProvider
        ksu = KensuProvider().instance()
        input_ds = KensuDatasourceAndSchema.for_path_with_opt_schema(ksu=ksu, ds_path=src, maybe_schema=maybe_schema, ds_name=src_name)
        output_ds = KensuDatasourceAndSchema.for_path_with_opt_schema(ksu=ksu, ds_path=dest, maybe_schema=maybe_schema, ds_name=dest_name)
        if maybe_schema is None:
            maybe_schema = [("unknown", "unknown")]
        lineage_info = [ExtDependencyEntry(
            input_ds=input_ds,
            # FIXME: kensu-py lineage do not work without schema as well...
            lineage= dict([(str(fieldname), [str(fieldname)])
                           for (fieldname, dtype) in maybe_schema]))]
        inputs_lineage = GenericComputedInMemDs(inputs=[input_ds], lineage=lineage_info)
        # register lineage in KensuProvider
        inputs_lineage.report(ksu=ksu, df_result=output_ds, operation_type=operation_type, report_output=True, register_output_orig_data=True)
        # actuly report the lineage and the write operation
        ksu.report_with_mapping()



from kensu.pandas import DataFrame,Series
from kensu.utils.kensu_provider import KensuProvider
import requests.models as md
from kensu.utils.helpers import eventually_report_in_mem
from kensu.numpy import ndarray

import builtins
# Extended subclass
class ksu_str(str):
    ksu_metadata = None

class Response(md.Response):

    ksu_schema = None
    ksu_short_schema = None
    ksu_stats = None

    @property
    def text(self) -> str:
        result = super(Response, self).text
        ksu_result = ksu_str(result)
        ksu_result.metadata = {"short_schema":self.ksu_short_schema, "real_schema":self.ksu_schema, "ds_location":self.ds_location, "stats":self.ksu_stats}
        return ksu_result



